# -*- coding: utf-8 -*-
from __future__ import absolute_import

from .base import stuf
from .core import defaultstuf, fixedstuf, frozenstuf, orderedstuf

idefaultstuf = defaultstuf
ifixedstuf = fixedstuf
ifrozenstuf = frozenstuf
iorderedstuf = orderedstuf
istuf = stuf
